#!/bin/bash


# TO DO:
#
# * Check exit status of encoder chain, user may have aborted.
#   in that case, do not start multiplexing and so on.
# 
# * Add processing profile for high quality XviD DVD backups
#
# * Check required diskspace for encoding + muxing + subtitle muxing
#
# * Add 'fast scaling' option to GUI, which uses cheap y4mscaler
#   settings.

# TESTING:
#
# * Audio/subtitle sync when converting AVI PAL -> NTSC





# How many frames/seconds to use for creating a preview
NumPreviewFrames=250
PreviewTime=10

# Read first 3 commandline arguments. 
# Command can be "Preview", "Probe" or empty.

InputFile="$1"
Command="$2"
PreviewPosition="$3"

echo "YUVDecoder is starting:" >> DebugOutput.txt
echo "YUVEncoder: Command is $Command, PreviewPosition is $PreviewPosition" >> DebugOutput.txt

# Check if other commandline arguments contain '--force-progressive'

if [ -n "`echo "$4$5$6$7" | grep -e '--force-progressive'`" ]; then

 ForceProgressiveInput="true"

 echo "OPTION: Forced progressive input." >> DebugOutput.txt

fi

# Check if other commandline arguments contain '--remove-overscan'

if [ -n "`echo "$4$5$6$7" | grep -e '--remove-overscan'`" ]; then

 RemoveOverscan="true"

 echo "OPTION: Remove overscan area from input." >> DebugOutput.txt

fi

# Report error and abort script

PrintError ()
{
 X=`type -p kdialog`

 if test $X && test -x $X && [ -n "$DISPLAY" ]; then

  kdialog --error "$1" 

 else

  echo >> DebugOutput.txt
  echo "---- RECEIVED ERROR MESSAGE ----" >> DebugOutput.txt
  echo >> DebugOutput.txt
  echo -e "$1" >> DebugOutput.txt
  echo >> DebugOutput.txt
  echo "--------------------------------" >> DebugOutput.txt

 fi

 exit 1
}

# Present a warning message to the user
# Argument $1 is warning message.

PrintWarning ()
{
 X=`type -p kdialog`

 if test $X && test -x $X && [ -n "$DISPLAY" ]; then
  kdialog --warningcontinuecancel "$1" >> DebugOutput.txt
  if [ ! "$?" -eq 0 ]; then
  echo >> DebugOutput.txt
  echo "USER ABORTED PROGRAM" >> DebugOutput.txt
  echo >> DebugOutput.txt
  exit 1
  fi
  
 else
  echo >> DebugOutput.txt
  echo "---- WARNING ----" >> DebugOutput.txt
  echo >> DebugOutput.txt
  echo -e "$1" >> DebugOutput.txt
  echo >> DebugOutput.txt
  echo "-----------------" >> DebugOutput.txt
 fi
}

# Detect FFMPEG version

DetectFFMPEG ()
{
    echo "$(ffmpeg 2>&1 | awk '$4 == "build" {print $5}' | sed s/,// | head -n 1)"
}

# Check if ffmpeg can decode video in file $1. 

DetectSupportedVideoCodec ()
{
 if [ -z "$(ffmpeg -i "$1" -f yuv4mpegpipe -pix_fmt yuv420p -t 1 -y /dev/null 2>&1 | grep 'Unsupported video codec')" ]; then

   echo "yes"

 fi
}

# Probe audio format of file $1 track $2
# $1 = audio input
# $2 = audio track

DetectAudioFormat ()
{
    echo "$(ffmpeg -i "$1" 2>&1 | \
    awk '/Audio:/ {print $4}' | sed s/,// | head -n ${2} | tail -n 1)"
}

# Probe video format of file $1
# $1 = video input

DetectVideoFormat ()
{
    echo "$(ffmpeg -i "$1" 2>&1 | \
    grep -o -E Video:\ .{1,\} | awk '{print $2}' | sed s/,// | head -n ${2} | tail -n 1)"
}


# Try if FFMPEG can decode audio in file $1
# $1 = video file

TestAudioDecoding ()
{

  ffmpeg -i "$1" -t 1 -f au -y /dev/null > /dev/null 2>&1

  if [ $? -ne 0 ]; then
    echo "false"
  else
    echo "true"
  fi

}

# Try if FFMPEG can encode audio from file $1 to ac3
# $1 = video file

TestAudioEncoding ()
{

  ffmpeg -i "$1" -t 1 -ab 384 -ar 48000 -ac 2 -f ac3 -y /dev/null > /dev/null 2>&1

  if [ $? -ne 0 ]; then
    echo "false"
  else
    echo "true"
  fi

}


# Probe interlacing of file $1. 
# Output is either 't', 'b' or 'p'

DetectVideoInterlacing ()
{

    ffmpeg -i "$1" -f yuv4mpegpipe -pix_fmt yuv420p -t 1 -y /tmp/tmp.y4m >/dev/null 2>&1
    echo "`head -n 1 /tmp/tmp.y4m | awk '{print $5}' | cut -c 2-`"
    rm -f /tmp/tmp.y4m
} 

# Probe framerate of file $1.
# Output is a ratio, like 25:1

DetectVideoFramerate ()
{
    ffmpeg -i "$1" -f yuv4mpegpipe -pix_fmt yuv420p -t 1 -y /tmp/tmp.y4m >/dev/null 2>&1
    echo "`head -n 1 /tmp/tmp.y4m | awk '{print $4}' | cut -c 2-`"
    rm -f /tmp/tmp.y4m
}

# Probe sample aspect ratio of file $1.
# Output is a ratio, like 59:54

DetectVideoSampleAspect ()
{
    ffmpeg -i "$1" -f yuv4mpegpipe -pix_fmt yuv420p -t 1 -y /tmp/tmp.y4m >/dev/null 2>&1
    echo "`head -n 1 /tmp/tmp.y4m | awk '{print $6}' | cut -c 2-`"
    rm -f /tmp/tmp.y4m
}

# Probe display aspect ratio of file $1, using Transcode.
# Output is a ratio, like 16:9

DetectVideoDisplayAspect ()
{

 X=`type -p tcprobe`
 if test $X && test -x $X; then

   DisplayAspect="$(tcprobe -i "$1" | grep 'aspect ratio' | grep -E -o [0-9]{1,\}:[0-9]{1,\})"

   if [ -z "$DisplayAspect" ]; then

    echo "Detection of video display aspect failed!" >> DebugOutput.txt
    echo ""

   else

   echo "$DisplayAspect"

   fi

 else

    echo "TCProbe was not found, no display aspect detected!" >> DebugOutput.txt
    echo ""

 fi

}

# Probe resolution of file $1.
# Output is of the form "640x480"

DetectVideoResolution ()
{
    ffmpeg -i "$1" -f yuv4mpegpipe -pix_fmt yuv420p -t 1 -y /tmp/tmp.y4m >/dev/null 2>&1
    echo "`head -n 1 /tmp/tmp.y4m | awk '{sub("W","",$2); sub("H","",$3); print $2"x"$3}'`"
    rm -f /tmp/tmp.y4m
}

# Detects length of file $1
# Outputs number of milliseconds

DetectVideoLength ()
{
  Hours=$(ffmpeg -i "$1" 2>&1 | grep Duration | awk '{print $2}' | cut -d : -f 1)
  Minutes=$(ffmpeg -i "$1" 2>&1 | grep Duration | awk '{print $2}' | cut -d : -f 2)
  Seconds=$(ffmpeg -i "$1" 2>&1 | grep Duration | awk '{print $2}' | cut -d : -f 3 | cut -d , -f 1)

  echo "$Hours*3600000 + $Minutes*60000 + $Seconds*1000" | bc | cut -d . -f 1 | cut -d , -f 1
}








# SCRIPT STARTS HERE

echo "Detecting file properties...." >> DebugOutput.txt

# Record ffmpeg output for debugging

echo "" >> DebugOutput.txt
echo "Output of ffmpeg -i:" >> DebugOutput.txt
ffmpeg -i "$InputFile" >> DebugOutput.txt 2>&1
echo "" >> DebugOutput.txt

# Do some tests first

ffmpeg_version="$(DetectFFMPEG)"
CanDecodeAudio=$(TestAudioDecoding "$InputFile")
CanEncodeAudio=$(TestAudioEncoding "$InputFile")
InputAudioFormat=$(DetectAudioFormat "$InputFile" 1)
InputVideoFormat=$(DetectVideoFormat "$InputFile" 1)

if [ $CanDecodeAudio == "false" ]; then

  if [ "$InputAudioFormat" == "ac3" ]; then

    echo "Will use stream-copy for ac3 soundtrack." >> DebugOutput.txt
    StreamCopy="yes"

  else

    PrintWarning "Your FFMPEG version is unable to decode the audio in this video file.\n\nWhen this file is converted to MPEG2 format, it will not contain any audio."
    EncodeSilence="yes"

  fi

fi

if [ $CanEncodeAudio == "false" ] && [ -z "$StreamCopy" ] && [ -z "$EncodeSilence" ]; then

  PrintError "Your FFMPEG version is unable to generate AC3 streams.\n\nAC3 is the audio format used on DVD disks. KmPg2 can not continue if your FFMPEG version can't generate audio in this format."

fi

VideoCodecSupported="$(DetectSupportedVideoCodec "$InputFile")"

if [ ! "$VideoCodecSupported" == "yes" ]; then

 PrintError "This video file requires one or more codecs that are not supported by your version of FFMPEG. Details can be found below.\n\n\n$(ffmpeg -i "$InputFile" 2>&1 | grep 'Stream')"

fi

# Detect file properties

InputVideoLength=$(DetectVideoLength "$InputFile")
InputInterlacing=$(DetectVideoInterlacing "$InputFile")
InputSampleAspect=$(DetectVideoSampleAspect "$InputFile")
InputFrameRate=$(DetectVideoFramerate "$InputFile")
InputResolution=$(DetectVideoResolution "$InputFile")

if [ -z "$InputResolution" ] || [ -z "$InputVideoLength" ]; then

 PrintError "Your version of FFMPEG failed to properly detect the properties of your video file.\n\nKmPg2 can not continue."

fi

InputXRes=$(echo $InputResolution | cut -d x -f 1)
InputYRes=$(echo $InputResolution | cut -d x -f 2)

if [ ! "$InputInterlacing" == "p" ] && [ "$ForceProgressive" == "true" ]; then

  InputInterlacing="p"

fi

# Calculate the number of frames in the input video

if [ -n "$InputVideoLength" ] && [ -n "$InputFrameRate" ]; then

  InputNumFrames=$[ ( $InputVideoLength * $(echo "$InputFrameRate" | cut -d : -f 1) ) / \
                    ( 1000 * $(echo "$InputFrameRate" | cut -d : -f 2) ) ]

else

  InputNumFrames=1
  echo "WARNING: Could not detect number of frames in input file" >> DebugOutput.txt

fi

echo "FFMPEG Probe: $InputSampleAspect" >> DebugOutput.txt

# Find or guess the input sample aspect when FFMPEG detection failed.
# Uses transcode.

InputSampleAspectX="$(echo "$InputSampleAspect" | cut -d : -f 1)"
InputSampleAspectY="$(echo "$InputSampleAspect" | cut -d : -f 2)"

if [ "$InputSampleAspectX" == "0" ] || [ "$InputSampleAspectY" == "0" ]; then

 # Input sample aspect was not detected.
 # We need to calculate it from the display aspect and resolution.

 InputDisplayAspect=$(DetectVideoDisplayAspect "$InputFile")

 echo "Transcode Probe: $(DetectVideoDisplayAspect "$InputFile")" >> DebugOutput.txt

 if [ -n "$InputDisplayAspect" ]; then

  InputDisplayAspectX="$(echo "$InputDisplayAspect" | cut -d : -f 1)"
  InputDisplayAspectY="$(echo "$InputDisplayAspect" | cut -d : -f 2)"

  InputSampleAspectX=$[ ( $InputYRes * $InputDisplayAspectX ) / $InputDisplayAspectY ]
  InputSampleAspectY=$InputYRes
  InputSampleAspect="${InputSampleAspectX}:${InputSampleAspectY}"

 else

   # Input display aspect not found. Assume square pixels.

   echo "ASSUMING 1:1 ASPECT!" >> DebugOutput.txt

   InputDisplayAspectX=$InputXRes
   InputDisplayAspectY=$InputYRes

   InputSampleAspectX=1
   InputSampleAspectY=1

   InputSampleAspect="1:1"
   InputDisplayAspect="${InputDisplayAspectX}:${InputDisplayAspectY}"

 fi

else

  # Sample aspect was detected without problems.

  InputSampleAspectX="$(echo "$InputSampleAspect" | cut -d : -f 1)"
  InputSampleAspectY="$(echo "$InputSampleAspect" | cut -d : -f 2)"

  InputDisplayAspectX=$[ $InputXRes * $InputSampleAspectX ]
  InputDisplayAspectY=$[ $InputYRes * $InputSampleAspectY ]
  InputDisplayAspect="${InputDisplayAspectX}:${InputDisplayAspectY}"

fi

# Determine the video norm of the input

if [ "$InputFrameRate" == "30000:1001" ]; then

   SourceVideoNorm="NTSC"

elif [ "$InputFrameRate" == "24000:1001" ]; then

   SourceVideoNorm="NTSC"

elif [ "$InputFrameRate" == "25:1" ]; then

   SourceVideoNorm="PAL"

elif [ "$InputFrameRate" == "2997:125" ]; then

   SourceVideoNorm="INVERSE-TELECINED-FILM"

else

   SourceVideoNorm="UNKNOWN"

fi

# Print some diagnostic info

echo "****************************************" >> DebugOutput.txt
echo "" >> DebugOutput.txt
echo "Detected ffmpeg version $ffmpeg_version." >> DebugOutput.txt
echo "Detected sox version `sox -h 2>&1 | grep Version`." >> DebugOutput.txt
echo "Video Norm is $SourceVideoNorm" >> DebugOutput.txt
echo "Video file interlacing is $InputInterlacing" >> DebugOutput.txt
echo "Video sample aspect is $InputSampleAspect" >> DebugOutput.txt
echo "Video display aspect is $InputDisplayAspect" >> DebugOutput.txt
echo "Frame rate is $InputFrameRate" >> DebugOutput.txt
echo "Playing time is $InputVideoLength milliseconds." >> DebugOutput.txt
echo "" >> DebugOutput.txt
echo "****************************************" >> DebugOutput.txt


if [ -n "$(which smil2raw)" ] && [ -z "$(smil2raw "$InputFile" -f 1 2>&1 | grep -o Error)" ]; then

 # Input file is a DV file AND SMILUtils is installed.
 # This means that we can create previews from material in
 # the middle of the video.

 InputIsDV="true"

 echo "Detected MiniDV file." >> DebugOutput.txt

 if [ "$Command" == "Preview" ]; then

   # We can seek in a DV file to create a preview render somewhere
   # in the middele of the stream, so find the requested position

   echo "Preview fragment requested, using smil2yuv for MiniDV decoding." >> DebugOutput.txt

   if [ -n "$PreviewPosition" ];  then

     FrameOffset=$(( ($PreviewPosition * $InputNumFrames) / 1000 ))

   else

     FrameOffset=0  

   fi

   # Check if the preview fragment is running past the end of the input video
   # and generate the decoding command.

   if [ $(( $FrameOffset + $NumPreviewFrames )) -ge $InputNumFrames ]; then

     FrameOffset=$(( $InputNumFrames - $NumPreviewFrames ))

     if [ $FrameOffset -lt 0 ]; then

      SMILFrameRange="-f $InputNumFrames"

     else 

      SMILFrameRange="-f $NumPreviewFrames -o $FrameOffset"

     fi

   else

     SMILFrameRange="-f $NumPreviewFrames -o $FrameOffset"

   fi

   echo "Using smil2yuv for preview MiniDV decoding. Decoding parameters: $SMILFrameRange." >> DebugOutput.txt

   # Test if smil2yuv is working

   smil2yuv "$InputFile" -f 1 > /dev/null

   if [ $? -ne 0 ]; then

     PrintError "The 'smil2yuv' utility from the SMILUtils package failed to read the video file.\n\nEither your MiniDV file is broken, or your SMILUtils version is broken."

   fi

   DecoderTool="smil2yuv"
   DecodeCommand="smil2yuv \"$InputFile\" $SMILFrameRange -i 2"

 else

   echo "Using FFMPEG for DV decoding." >> DebugOutput.txt

   DecoderTool="FFMPEG"

   # generate command that decodes the input video into yuv4mpeg format on stdout.

   if [ -z "$ffmpeg_version" ]; then

     # We have CVS version

     DecodeCommand="ffmpeg -i \"$InputFile\" -f image2pipe -vcodec pgmyuv -y /dev/stdout | pgmtoy4m -r $InputFrameRate -i $InputInterlacing -a $InputSampleAspect "

   else

     if test "$ffmpeg_version" -ge "4731"; then

       DecodeCommand="ffmpeg -i \"$InputFile\" -f image2pipe -vcodec pgmyuv -y /dev/stdout | pgmtoy4m -r $InputFrameRate -i $InputInterlacing -a $InputSampleAspect "

     else

       DecodeCommand="ffmpeg -i \"$InputFile\" -f imagepipe -img pgmyuv -y /dev/stdout | pgmtoy4m -r $InputFrameRate -i $InputInterlacing -a $InputSampleAspect "

     fi

   fi

 fi


else

  # Input file is NOT a DV file.
  # This means we need to use FFMPEG and can't seek.

  InputIsDV="false"

  DecoderTool="FFMPEG"

  echo "Input file is not MiniDV, using FFMPEG for decoding." >> DebugOutput.txt

  if [ "$Command" == "Preview" ]; then

    echo "COMMAND is PREVIEW." >> DebugOutput.txt

    FFMPEGFrameLimit="-t $PreviewTime"

  fi

  # generate command that decodes the input video into yuv4mpeg format on stdout.

   if [ -z "$ffmpeg_version" ]; then

     # We have CVS version

     DecodeCommand="ffmpeg -i \"$InputFile\" $FFMPEGFrameLimit -f image2pipe -vcodec pgmyuv -y /dev/stdout | pgmtoy4m -r $InputFrameRate -i $InputInterlacing -a $InputSampleAspect "

   else

     if test "$ffmpeg_version" -ge "4731"; then

       DecodeCommand="ffmpeg -i \"$InputFile\" $FFMPEGFrameLimit -f image2pipe -vcodec pgmyuv -y /dev/stdout | pgmtoy4m -r $InputFrameRate -i $InputInterlacing -a $InputSampleAspect "

     else

       DecodeCommand="ffmpeg -i \"$InputFile\" $FFMPEGFrameLimit -f imagepipe -img pgmyuv -y /dev/stdout | pgmtoy4m -r $InputFrameRate -i $InputInterlacing -a $InputSampleAspect "

     fi

  fi

fi

# Change yuv4mpeg header to progressive when needed.

if [ "$ForceProgressiveInput" == "true" ]; then

  echo "Forcing progressive output." >> DebugOutput.txt

  DecodeCommand="$DecodeCommand | yuvcorrect -T PROGRESSIVE"

fi

# Cut the edges off the input video scale it such that the resolution
# remains the same.

if [ "$RemoveOverscan" == "true" ]; then

  ReducedXRes=$[ ( $InputXRes * 90 ) / 100 ]
  ReducedYRes=$[ ( $InputYRes * 90 ) / 100 ]

  echo "Removing overscan area." >> DebugOutput.txt

  RemoveOverscanCommand="-I active=${ReducedXRes}x${ReducedYRes}+0+0cc -O active=${InputXRes}x${InputYRes}+0+0cc"

fi

# This is what the final command looks like.

DecodeCommand="$DecodeCommand | y4mscaler $RemoveOverscanCommand -O chromass=420_MPEG2 -S option=sinc:4"

if [ "$Command" == "Probe" ]; then

  echo "COMMAND is PROBE." >> DebugOutput.txt


  echo -e "FrameRate\t$InputFrameRate"
  echo -e "SampleAspect\t$InputSampleAspect"
  echo -e "Resolution\t$InputResolution"
  echo -e "Norm\t$SourceVideoNorm"
  echo -e "AudioFormat\t$InputAudioFormat"
  echo -e "VideoFormat\t$InputVideoFormat"

  if [ "$InputVideoLength" -gt 7200000 ]; then

    echo -e "Length\t$(( $InputVideoLength / 3600000)) Hours"

  elif [ "$InputVideoLength" -gt 120000 ]; then

    echo -e "Length\t$(( $InputVideoLength / 60000)) Minutes"

  elif [ "$InputVideoLength" -gt 2000 ]; then

    echo -e "Length\t$(( $InputVideoLength / 1000)) Seconds"

  else 

    echo -e "Length\t$InputVideoLength Milliseconds"

  fi

  if [ "$InputInterlacing" == "p" ]; then

    echo -e "Interlacing\tNo"

  else

    echo -e "Interlacing\tYes ($InputInterlacing)"

  fi

  echo -e "CanSeek\t$InputIsDV"

  exit

fi

# Pass on info to next script

echo "$InputFrameRate"
echo "$InputSampleAspect"
echo "$InputXRes"
echo "$InputYRes"
echo "$InputSampleAspectX"
echo "$InputSampleAspectY"
echo "$InputVideoLength"

# Generate YUV Stream on stdout:

echo "YUVDecoder COMMAND: $DecodeCommand" >> DebugOutput.txt

eval "$DecodeCommand"